/**
 * Copyright 2023 actions-toolkit authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
export declare const setupDockerWinPs1: () => string;
export declare const dockerServiceLogsPs1: () => string;
export declare const colimaYaml: () => string;
export declare const setupDockerWinPs1Data = "\n[CmdletBinding()]\nparam(\n    [Parameter(Mandatory = $true)]\n    [string]$ToolDir,\n\n    [Parameter(Mandatory = $true)]\n    [string]$RunDir,\n\n    [Parameter(Mandatory = $true)]\n    [string]$DockerHost,\n\n    [Parameter(Mandatory = $false)]\n    [string]$DaemonConfig)\n\n$pwver = (Get-ItemProperty -Path HKLM:\\SOFTWARE\\Microsoft\\PowerShell\\3\\PowerShellEngine -Name 'PowerShellVersion').PowerShellVersion\nWrite-Host \"PowerShell version: $pwver\"\n\n# Create run directory\nNew-Item -ItemType Directory \"$RunDir\" -ErrorAction SilentlyContinue | Out-Null\n\n# Remove existing service\nif (Get-Service docker -ErrorAction SilentlyContinue) {\n  $dockerVersion = (docker version -f \"{{.Server.Version}}\")\n  Write-Host \"Current installed Docker version: $dockerVersion\"\n  # stop service\n  Stop-Service -Force -Name docker\n  Write-Host \"Service stopped\"\n  # remove service\n  sc.exe delete \"docker\"\n  # removes event log entry. we could use \"Remove-EventLog -LogName -Source docker\"\n  # but this cmd is not available atm\n  $ErrorActionPreference = \"SilentlyContinue\"\n  & reg delete \"HKLM\\SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\docker\" /f 2>&1 | Out-Null\n  $ErrorActionPreference = \"Stop\"\n  Write-Host \"Service removed\"\n}\n\n$env:DOCKER_HOST = $DockerHost\nWrite-Host \"DOCKER_HOST: $env:DOCKER_HOST\"\n\nif ($DaemonConfig) {\n  Write-Host \"Writing Docker daemon config\"\n  New-Item -ItemType Directory -Force -Path \"$env:ProgramData\\Docker\\config\"\n  $DaemonConfig | Out-File -FilePath \"$env:ProgramData\\Docker\\config\\daemon.json\"\n}\n\nWrite-Host \"Creating service\"\nNew-Item -ItemType Directory \"$RunDir\\moby-root\" -ErrorAction SilentlyContinue | Out-Null\nNew-Item -ItemType Directory \"$RunDir\\moby-exec\" -ErrorAction SilentlyContinue | Out-Null\nStart-Process -Wait -NoNewWindow \"$ToolDir\\dockerd\" `\n  -ArgumentList `\n    \"--host=$DockerHost\", `\n    \"--data-root=$RunDir\\moby-root\", `\n    \"--exec-root=$RunDir\\moby-exec\", `\n    \"--pidfile=$RunDir\\docker.pid\", `\n    \"--register-service\"\nWrite-Host \"Starting service\"\nStart-Service -Name docker\nWrite-Host \"Service started successfully!\"\n\n$tries=20\nWrite-Host \"Waiting for Docker daemon to start...\"\nWhile ($true) {\n  $ErrorActionPreference = \"SilentlyContinue\"\n  & \"$ToolDir\\docker\" version | Out-Null\n  $ErrorActionPreference = \"Stop\"\n  If ($LastExitCode -eq 0) {\n    break\n  }\n  $tries--\n  If ($tries -le 0) {\n    Throw \"Failed to get a response from Docker daemon\"\n  }\n  Write-Host -NoNewline \".\"\n  Start-Sleep -Seconds 1\n}\nWrite-Host \"Docker daemon started successfully!\"\n";
export declare const dockerServiceLogsPs1Data = "\nGet-WinEvent -ea SilentlyContinue `\n  -FilterHashtable @{ProviderName= \"docker\"; LogName = \"application\"} |\n    Sort-Object @{Expression=\"TimeCreated\";Descending=$false} |\n    ForEach-Object {\"$($_.TimeCreated.ToUniversalTime().ToString(\"o\")) [$($_.LevelDisplayName)] $($_.Message)\"}\n";
export declare const colimaYamlData = "\n# Number of CPUs to be allocated to the virtual machine.\n# Default: 2\ncpu: 2\n\n# Size of the disk in GiB to be allocated to the virtual machine.\n# NOTE: changing this has no effect after the virtual machine has been created.\n# Default: 60\ndisk: 60\n\n# Size of the memory in GiB to be allocated to the virtual machine.\n# Default: 2\nmemory: 2\n\n# Architecture of the virtual machine (x86_64, aarch64, host).\n# Default: host\narch: host\n\n# Container runtime to be used (docker, containerd).\n# Default: docker\nruntime: docker\n\n# Kubernetes configuration for the virtual machine.\nkubernetes:\n  enabled: false\n\n# Auto-activate on the Host for client access.\n# Setting to true does the following on startup\n#  - sets as active Docker context (for Docker runtime).\n#  - sets as active Kubernetes context (if Kubernetes is enabled).\n# Default: true\nautoActivate: false\n\n# Network configurations for the virtual machine.\nnetwork:\n  # Assign reachable IP address to the virtual machine.\n  # NOTE: this is currently macOS only and ignored on Linux.\n  # Default: false\n  address: false\n\n  # Custom DNS resolvers for the virtual machine.\n  #\n  # EXAMPLE\n  # dns: [8.8.8.8, 1.1.1.1]\n  #\n  # Default: []\n  dns: []\n\n  # DNS hostnames to resolve to custom targets using the internal resolver.\n  # This setting has no effect if a custom DNS resolver list is supplied above.\n  # It does not configure the /etc/hosts files of any machine or container.\n  # The value can be an IP address or another host.\n  #\n  # EXAMPLE\n  # dnsHosts:\n  #   example.com: 1.2.3.4\n  dnsHosts:\n    host.docker.internal: host.lima.internal\n\n  # Network driver to use (slirp, gvproxy), (requires vmType `qemu`)\n  #   - slirp is the default user mode networking provided by Qemu\n  #   - gvproxy is an alternative to VPNKit based on gVisor https://github.com/containers/gvisor-tap-vsock\n  # Default: gvproxy\n  driver: gvproxy\n\n# Forward the host's SSH agent to the virtual machine.\n# Default: false\nforwardAgent: false\n\n# Docker daemon configuration that maps directly to daemon.json.\n# https://docs.docker.com/engine/reference/commandline/dockerd/#daemon-configuration-file.\n# NOTE: some settings may affect Colima's ability to start docker. e.g. `hosts`.\n#\n# EXAMPLE - disable buildkit\n# docker:\n#   features:\n#     buildkit: false\n#\n# EXAMPLE - add insecure registries\n# docker:\n#   insecure-registries:\n#     - myregistry.com:5000\n#     - host.docker.internal:5000\n#\n# Colima default behaviour: buildkit enabled\n# Default: {}\n{{daemonConfig}}\n\n# Virtual Machine type (qemu, vz)\n# NOTE: this is macOS 13 only. For Linux and macOS <13.0, qemu is always used.\n#\n# vz is macOS virtualization framework and requires macOS 13\n#\n# Default: qemu\nvmType: qemu\n\n# Volume mount driver for the virtual machine (virtiofs, 9p, sshfs).\n#\n# virtiofs is limited to macOS and vmType `vz`. It is the fastest of the options.\n#\n# 9p is the recommended and the most stable option for vmType `qemu`.\n#\n# sshfs is faster than 9p but the least reliable of the options (when there are lots\n# of concurrent reads or writes).\n#\n# Default: virtiofs (for vz), sshfs (for qemu)\nmountType: 9p\n\n# The CPU type for the virtual machine (requires vmType `qemu`).\n# Options available for host emulation can be checked with: `qemu-system-$(arch) -cpu help`.\n# Instructions are also supported by appending to the cpu type e.g. \"qemu64,+ssse3\".\n# Default: host\ncpuType: host\n\n# For a more general purpose virtual machine, Ubuntu container is optionally provided\n# as a layer on the virtual machine.\n# The underlying virtual machine is still accessible via `colima ssh --layer=false` or running `colima` in\n# the Ubuntu session.\n#\n# Default: false\nlayer: false\n\n# Custom provision scripts for the virtual machine.\n# Provisioning scripts are executed on startup and therefore needs to be idempotent.\n#\n# EXAMPLE - script exected as root\n# provision:\n#   - mode: system\n#     script: apk add htop vim\n#\n# EXAMPLE - script exected as user\n# provision:\n#   - mode: user\n#     script: |\n#       [ -f ~/.provision ] && exit 0;\n#       echo provisioning as $USER...\n#       touch ~/.provision\n#\n# Default: []\nprovision:\n  - mode: system\n    script: |\n      mkdir -p /tmp/docker-bins\n      cd /tmp/docker-bins\n      wget -qO- \"https://download.docker.com/linux/static/{{dockerBinChannel}}/{{dockerBinArch}}/docker-{{dockerBinVersion}}.tgz\" | tar xvz --strip 1\n      mv -f /tmp/docker-bins/* /usr/bin/\n\n# Modify ~/.ssh/config automatically to include a SSH config for the virtual machine.\n# SSH config will still be generated in ~/.colima/ssh_config regardless.\n# Default: true\nsshConfig: false\n\n# Configure volume mounts for the virtual machine.\n# Colima mounts user's home directory by default to provide a familiar\n# user experience.\n#\n# EXAMPLE\n# mounts:\n#   - location: ~/secrets\n#     writable: false\n#   - location: ~/projects\n#     writable: true\n#\n# Colima default behaviour: $HOME and /tmp/colima are mounted as writable.\n# Default: []\nmounts: []\n\n# Environment variables for the virtual machine.\n#\n# EXAMPLE\n# env:\n#   KEY: value\n#   ANOTHER_KEY: another value\n#\n# Default: {}\nenv: {}\n";
export declare const qemuEntitlements = "\n<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">\n<plist version=\"1.0\">\n<dict>\n    <key>com.apple.security.hypervisor</key>\n    <true/>\n</dict>\n</plist>\n";
